/* kjis.c -- Module for drawing 16x16 Kanji's, font file I/O.
 *	Support of complete JIS level-1 and -2 font (6877 chars).
 *	Three modes for JIS level-2 font access configurable via KD.INI:
 *	    NONE -- level-2 font file will not be used at all.
 *	    File -- level-2 font file is accessed as needed on the fly.
 *	    RAM  -- Both level-1 and -2 fonts are loaded into RAM at startup.
 *		Izumi Ohzawa, 2-16-90
 *
 *      Modified for speed by Koichi Mori, Caltech [Jan 25 90]
 *
 *      Re-written in C (Turbo C 2.0) by:
 *		Izumi Ohzawa, 1-8-90
 *		Group in Neurobiology/School of Optometry
 *		University of California, Berkeley
 *		Berkeley, CA 94720
 *		izumi@violet.berkeley.edu
 *
 *	This program/code is hereby released into the public domain.
 *	No copyright is claimed, and no restriction is imposed for any use,
 *	commercial or otherwise.  Use at your own risk.
 * --------------------------------------------------------------------------
 *
 *	Ported from JIS.PAS (Pascal) which is a module in a program called
 *	READER (JIS Kanji file browser).  JIS.PAS is by (6-11-1989):
 *		Seiichi Nomura, Associate Professor of Mechanical Engineering
 *		University of Texas, Arlington
 *		Arlington, Texas 76019
 *		B470SSN@UTARLVM1.BITNET, B460NOM@UTARLG.ARL.UTEXAS.EDU
 *
 * 	        and by (12-24-1989)
 *		Seke Wei
 *		Institute of Computer Science
 *		TsingHua University
 *		HsinChu, Taiwan 30043, ROC
 *		TH774339@TWNCTU01.BITNET
 * ------------------------------------------------------------------------
 * Function OutKanji() is based on "SJIS.C" by:
 * 		Mark Edwards, [Jan. 1990]
 *		University of Wisconsin-Madison.
 * ------------------------------------------------------------------------- */

#define GLOB_ALLOC 1		/* Allocate storage for global vars */

#include <stdio.h>
#include <dos.h>
#include <stdlib.h>
#include <alloc.h>
#include <graphics.h>
#include <string.h>
#include <io.h>
#include "kjis.h"
#include "prototyp.h"

#define KANJI_INVALID	3489	/* Must be >2000, or fix Level2 file read code */
				/* Index into JIS-1 font for invalid font */

static char *JIS1_file = "\\k16jis1.fnt";	/* JIS level-1 font file */
static char *JIS2_file = "\\k16jis2.fnt";	/* JIS level-2 font file */
static FILE *fp;
Fontpack *fontptr1;                /* first 64 kbytes (32*2000 chrs) of JIS-1 */
Fontpack *fontptr2;                /* rest (32*(1489+1) chars) of JIS-1 */
Fontpack *fontptr3;		   /* RAM mode first 64kb of JIS-2 */
Fontpack *fontptr4;		   /* RAM mode rest of JIS-2 */
static int *cache_list;   	   /* array for index of chars in cache */
static int Ncache = 0;		   /* next cache location */
static int NNcache = 0;		   /* number of chars in cache. for() loop end cond */


/* === Function to check if a level-2 kanji is already in cache ======== */
int
Check_Cache(int kindex)
{
int jj, retval;
	retval = -1;
	for(jj=0; jj<NNcache; jj++)
	   if(kindex == cache_list[jj])
	   {
		retval = jj;
		break;
	   }
	return(retval);	
}


/* === Function to read in JIS Kanji font into RAM ====================== */
int
Read_kjfont(char *fontdir)
{
int count1, count2, iii;
char fontfilename[128];

/* #### JIS level-1 font (3489 charsi +1) will be loaded into RAM */
	strcpy(fontfilename, fontdir);	
	strcat(fontfilename, JIS1_file);
        if( (fp = fopen(fontfilename,"rb")) == NULL)
        {
		fprintf(stderr,"Can't open JIS-1 font file: %s\n",fontfilename);
                return(1);
        }
        fontptr1 = (Fontpack *)malloc(2000*sizeof(Fontpack));
        if(fontptr1 == NULL)
        {
                fprintf(stderr,"malloc() for fontptr1 failed.\n");
                fclose(fp);
                return(2);
        }
        fontptr2 = (Fontpack *)malloc(1490*sizeof(Fontpack));
        if(fontptr2 == NULL)
        {
                fprintf(stderr,"malloc() for fontptr2 failed.\n");
                free(fontptr1);
                fclose(fp);
                return(3);
        }

        fprintf(stderr,"Array allocated... Reading JIS-1 Font... ");
        count1 = fread( (void *)fontptr1, sizeof(Fontpack), 2000, fp);
        count2 = fread( (void *)fontptr2, sizeof(Fontpack), 1490, fp);
        if(count1 != 2000 || count2 != 1490)
        {
                fprintf(stderr, "\nJIS-1 read failure -- %d %d\n", count1, count2);
                free(fontptr2);
                free(fontptr1);
                fclose(fp);
                return(4);
        }
        fclose(fp);
        fprintf(stderr,"DONE\n");

/* #### Now deal with JIS level-2 font differently depending on Level2Font */
/*	There are 3388 level-2 chars in file */

	if(Level2Font > 0)	/* If RAM or File */
	{
	    strcpy(fontfilename, fontdir);	
	    strcat(fontfilename, JIS2_file);
            if( (fp = fopen(fontfilename,"rb")) == NULL)
            {
		fprintf(stderr,"Can't open JIS-2 font file: %s\n",fontfilename);
		fprintf(stderr,"Only JIS level-1 is available.\n");
                Level2Font = 0;
            }
	} /* end of if(Level2Font > 0) */

	if(Level2Font == JIS2_RAM)	/* JIS level-2 RAM Font */
	{
            fontptr3 = (Fontpack *)malloc(2000*sizeof(Fontpack));
            if(fontptr3 == NULL)
            {
                fprintf(stderr,"malloc() for fontptr3 failed.\n");
                free(fontptr2);
                free(fontptr1);
                fclose(fp);
                return(6);
            }
            fontptr4 = (Fontpack *)malloc(1388*sizeof(Fontpack));
            if(fontptr4 == NULL)
            {
                fprintf(stderr,"malloc() for fontptr4 failed.\n");
                free(fontptr3);
                free(fontptr2);
                free(fontptr1);
                fclose(fp);
                return(7);
            }

            fprintf(stderr,"Array allocated... Reading JIS-2 Font... ");
            count1 = fread( (void *)fontptr3, sizeof(Fontpack), 2000, fp);
            count2 = fread( (void *)fontptr4, sizeof(Fontpack), 1388, fp);
            if(count1 != 2000 || count2 != 1388)
            {
                fprintf(stderr, "\nJIS-2 read failure -- %d %d\n", count1, count2);
		free(fontptr4);
                free(fontptr3);
                free(fontptr2);
                free(fontptr1);
                fclose(fp);
                return(8);
            }
            fprintf(stderr,"DONE\n");
            fclose(fp);
	}  /* end of if(Level2Font == JIS2_RAM) .. */
	else if(Level2Font == JIS2_FILE)
	{ 
            fontptr3 = (Fontpack *)malloc(SIZECACHE*sizeof(Fontpack));
	    cache_list = (int *)malloc(SIZECACHE*sizeof(int));
            if(fontptr3 == NULL || cache_list == NULL)
            {
                fprintf(stderr,"malloc() for font cache failed.\n");
		Level2Font = 0;
            }
	    for(iii=0; iii<SIZECACHE; iii++)
		cache_list[iii] = 0;		/* clear cache */
	} /* leave the JIS-2 font file open */
        return(0);
}

/* === Function to convert JIS Kanji code to index into our font data */
int
jis2index(unsigned int jiscode)
{
int hi, lo;
	hi = jiscode >>8;
	lo = jiscode & 0x00ff;

	if(lo>0x20 && lo<0x7F && hi>0x20 && hi<0x74)
            hi = 94*hi + lo - 3135;  /* hi=94*(hi-33)+(lo-33) */
	else
            hi = 0x7e7e;   /* debug by K.M. *//* code out of range, apple mark */
	/* "hi" is reused */

        /*...........................range....notation..JIS code.....index... */
	if(hi>=282 && hi <=364)
		hi -= 73; 	/* 282..364  hiragana, $2421-$2473, 209-291  */
	else if(hi>=1410 && hi<=4374)
		hi -= 886; 	/* 1410..4374 kanji-1, $3021-$4f53, 524-3488 */
	else if(hi>=376 && hi<=461)
		hi -= 84; 	/* 376..461  katakana, $2521-$2576, 292-377  */
	else if(hi>=203 && hi<=212)
		hi -= 56;  	/* 203..212   0 .. 9,  $2330-$2339, 147-156  */
	else if(hi>=220 && hi<=245)
		hi -= 63; 	/* 220..245   A .. Z,  $2341-$235a, 157-182  */
	else if(hi>=252 && hi<=277)
		hi -= 69; 	/* 252..277   a .. z,  $2361-$237a, 183-208  */
	else if(hi>=470 && hi<=493)
                hi -= 92;      	/* 470..493   GREEK,   $2621-$2638  378-401  */
	else if(hi>=502 && hi<=525)
                hi -= 100;     	/* 502..525   greek,   $2641-$2658  402-425  */
	else if(hi>=564	&& hi<=596)
                hi -= 138;     	/* 564..596   RUSSIAN, $2721-$2741  426-458  */
	else if(hi>=612 && hi<=644)
                hi -= 153;     	/* 612..644   russian, $2751-$2771  459-491  */
				/*  n/a       linedraw,  n/a        492-523  */
	else if(hi<= 107)
		;  		/* 0..93               $2121-$217e, 0  -93   */
			   	/* 94..107             $2221-$222e, 94 -107  */
	else if(Level2Font>0 && hi>=4418 && hi<=7805)
		hi -= 928;      /* 4418..7805 kanji-2, $5021-$737e, 3490-6877 */
	else	hi = KANJI_INVALID;   /* non printable */
	return(hi);
}


/* === Function to display 1 Kanji at current position ================ */
void
OutKanji(int index)     /* 0 <= index <= 6876 */
{
register int LRbits, ix;
static int *font;
static int x,y,iy;
static errcode1, errcode2;
static long j2offset;	/* for fseek() */

        /* Read current position */
        x = Xcur<<3;		/* Xcur*8 */
	y = Ycur * PixelPerRow;

        if(index < 2000)
                font = (int *)&fontptr1[index];		/* JIS-1 lower half */
        else if(index < 3490)
                font = (int *)&fontptr2[index-2000];	/* JIS-1 upper half */
	else if(Level2Font == JIS2_RAM)			/* JIS-2 RAM font */
	{
	    if(index < 5490)
		font = (int *)&fontptr3[index-3490];	/* JIS-2 lower half */
	    else
		font = (int *)&fontptr4[index-5490];	/* JIS-2 upper half */
	}
	else if(Level2Font == JIS2_FILE)		/* JIS-2 Disk File font */
	{
	    /* -1 if not in cache, cache index if found */
	    ix = Check_Cache(index);
	    if(ix >= 0)		/* found in cache */
		font = (int *)&fontptr3[ix];
	    else		/* not in cache, get bitmap from file */
	    {
		/* This is where DOS is RE-ENTERED.  Because of this	*/
		/* re-entry, KD may fail when "Level2Font File" is      */
		/* specified in the initialization file "kd.ini"	*/
	        j2offset = index - 3490;
	        j2offset *= sizeof(Fontpack);   /* offset into JIS-2 file */
	        errcode1 = fseek(fp, j2offset, SEEK_SET);
	        errcode2 = fread((void *)&fontptr3[Ncache], sizeof(Fontpack), 1, fp);
		if(errcode1 == 0 && errcode2 == 1)	/* Char read OK */
		{
	            font = (int *)&fontptr3[Ncache];
		    cache_list[Ncache] = index;	/* store entry in list */
		    Ncache++;			/* point to next location */
		    Ncache %= SIZECACHE;	/* circular cache.  Oldest scrolls off */
		    NNcache++;			/* for() loop end number */
		    if(NNcache > SIZECACHE)
		        NNcache = SIZECACHE;	/* limit there */
		}
		else	/* This level-2 char couldn't be read */
		    font = (int *)&fontptr2[KANJI_INVALID - 2000];
	    }
	} /* end of else if(Level2Font == JIS2_FILE) .. */

	setfillstyle(SOLID_FILL, 0);
	bar(x, y, x+15, y+15);
        /* display kanji -- font stored in TV scan format L->R, Top->Bottom */
	for(iy=0; iy< 16; iy++)
	{
                LRbits = font[iy];
		for(ix=0; ix< 16; ix++)
		{
		    if (LRbits & 1)
		        putpixel(ix+x, iy+y, KanjiColor);
		    LRbits >>= 1;
		}
	}
        Xcur += 2;        /* Move 2 ASCII widths for next kanji */
}


/* === Function to display one ASCII or Kanji character */
void
Show_Char(unsigned int ascjis)
{
int xpos, ypos;
static char chascii[] = " ";
	if(ascjis & 0xff00)	/* if hi byte is non-zero */
	    OutKanji(jis2index(ascjis));
	else			/* Normal ASCII if hi byte == 0 */
	{
	    *chascii = ascjis;
	    xpos = Xcur<<3;		/* Could use Toggle_Cursor() here and at bottom */
	    ypos = Ycur*PixelPerRow;	/* but maybe this is a bit faster. */
	    putimage(lastxpos, lastypos, Cursor_Bitmap, XOR_PUT); /* take off curs */
	    setfillstyle(SOLID_FILL, AscBackColor);
	    bar(xpos, ypos, xpos+7, ypos+15);	/* clear out background */
	    setcolor(AscColor);
	    outtextxy(xpos, ypos+4, chascii);
	    if(Xcur < MaxColumn)
	       Xcur++;
	    lastxpos = xpos = Xcur<<3;
	    lastypos = ypos+PixelPerRow-1;
	    putimage(lastxpos, lastypos, Cursor_Bitmap, XOR_PUT);	/* place cursor */
	}
}


